const { Pool } = require('pg');

const dbConfig = {
    user: process.env.POSTGRES_USER || 'e2ee',
    password: process.env.POSTGRES_PASSWORD,
    host: process.env.POSTGRES_HOST || '127.0.0.1',
    port: process.env.POSTGRES_PORT || '5432',
    database: process.env.POSTGRES_DB || 'e2ee',
};

const pool = new Pool(dbConfig);

const database = {
    init: async () => {
        while (true) {
            try {
                await database.checkIfDatabaseIsUp();
                console.log('Database is up, initializing...');
                database.checkSchema().then((exists) => {
                    if (!exists) {
                        database.createTables();
                    } else {
                        console.log("Database schema is already correct.");
                    }
                }).catch((err) => {
                    console.error("Error checking schema:", err);
                });
                break;
            } catch (err) {
                console.error('Waiting for the database to start...');
            }
            await new Promise(resolve => setTimeout(resolve, 5000));
        }
    },
    createTables: () => {
        pool.query(`
            CREATE TABLE IF NOT EXISTS "users" (
                uuid SERIAL PRIMARY KEY,
                pubkey text
            );
            CREATE TABLE IF NOT EXISTS "room" (
                    uuid SERIAL PRIMARY KEY
            );
            CREATE TABLE IF NOT EXISTS "room_members" (
                room_uuid INTEGER REFERENCES "room"(uuid),
                user_uuid INTEGER REFERENCES "users"(uuid),
                PRIMARY KEY (room_uuid, user_uuid)
            );
        `, (err, _) => {
            if (err) {
                console.error('Error creating tables', err);
                return;
            } else {
                    console.log("tables created successfully.");
                }
        });
    },
    checkSchema: () => {
        return new Promise((resolve, reject) => {
            pool.query(`
                SELECT
                    (EXISTS (SELECT 1 FROM pg_tables WHERE tablename = 'users'))
                    AND
                    (EXISTS (SELECT 1 FROM pg_tables WHERE tablename = 'room'))
                    AND
                    (EXISTS (SELECT 1 FROM pg_tables WHERE tablename = 'room_members'))
                AS all_tables_exist;
            `, (err, res) => {
                if (err) {
                    console.error('Error executing query', err);
                    reject(err);
                } else {
                    const all_tables_exist = res.rows[0].all_tables_exist;
                    resolve(all_tables_exist);
                }
            });
        });
    },
    checkIfDatabaseIsUp: () => {
        return new Promise((resolve, reject) => {
            pool.query('SELECT 1', (err, res) => {
                if (err) {
                    reject(false);
                } else {
                    resolve(true);
                }
            });
        });
    },
    addUser: async (pubkey) => {
        if (!pubkey) {
            console.error("Pubkey is required");
            return;
        }
        await retryWithDelay(database.checkIfDatabaseIsUp, 5);
        try {
            await pool.query(
                'INSERT INTO "users" (uuid, pubkey) VALUES (DEFAULT, $1)',
                [pubkey]
            );
            console.log(`Added user with the public key ${pubkey}.`);
            return;
        } catch (err) {
            console.error('Error adding user:', err);
            throw err;
        }
    },
    createRoom: async (pubkey1, pubkey2) => {
        await retryWithDelay(database.checkIfDatabaseIsUp, 5);
        try {
            const userQuery = 'SELECT uuid FROM users WHERE pubkey = $1';
            const uuidRes1 = await pool.query(userQuery, [pubkey1]);
            const uuidRes2 = await pool.query(userQuery, [pubkey2]);

            if (!uuidRes1.rows[0] || !uuidRes2.rows[0]) {
                throw new Error('One or both users not found');
            }

            const uuid1 = uuidRes1.rows[0].uuid;
            const uuid2 = uuidRes2.rows[0].uuid;

            const roomRes = await pool.query(
                'INSERT INTO room (uuid) VALUES (DEFAULT) RETURNING uuid'
            );
            const roomid = roomRes.rows[0].uuid;

            await pool.query(
                `INSERT INTO room_members (room_uuid, user_uuid)
                VALUES ($1, $2), ($1, $3)`,
                [roomid, uuid1, uuid2]
            );
            return roomid;
        } catch (err) {
            console.error('Error creating the room:', err);
            throw err;
        }
    },
    getRooms: async (pubkey) => {
    await retryWithDelay(database.checkIfDatabaseIsUp, 5);
    try {
        const roomsRes = await pool.query(
            `SELECT m.room_uuid, u.pubkey
             FROM room_members m, users u
             WHERE m.user_uuid = u.uuid
             AND m.room_uuid in (
                SELECT room_uuid
                FROM room_members
                WHERE user_uuid = (SELECT uuid FROM users WHERE pubkey = $1)
             ) AND u.pubkey != $1`,
            [pubkey]
        );
        return roomsRes.rows.map(row => [row.room_uuid, row.pubkey]);
    } catch (err) {
        console.error('Error retrieving rooms:', err);
        throw err;
    }
},

    getRoomMembers: async (roomid) => {
        await retryWithDelay(database.checkIfDatabaseIsUp, 5);
        try {
            const members = await pool.query(
                'SELECT u.pubkey FROM room_members r, users u WHERE r.user_uuid = u.uuid AND r.room_uuid = $1',
                [roomid]
            );
            return members.rows.map(row => row.pubkey);
        } catch (err) {
            console.error('Error retrieving rooms:', err);
            throw err;
        }
    },
    getPeers: async (pubkey) => {
        await retryWithDelay(database.checkIfDatabaseIsUp, 5);
        try {
            const peers = await pool.query(
                `SELECT u1.pubkey
                FROM room_members r1, room_members r2, users u1, users u2
                WHERE r1.user_uuid = u1.uuid
                AND r2.user_uuid = u2.uuid
                AND r1.room_uuid = r2.room_uuid
                AND u2.pubkey != u1.pubkey
                AND u2.pubkey = $1`,
                [pubkey]
            );
            return peers.rows.map(row => row.pubkey);
        } catch (err) {
            console.error('Error retrieving peers:', err);
            throw err;
        }
    },
    getPublicKeys: async () => {
        await retryWithDelay(database.checkIfDatabaseIsUp, 5);
        try {
            const result = await pool.query('SELECT pubkey FROM users');
            const publicKeys = result.rows.map(row => row.pubkey);
            return publicKeys;
        } catch (err) {
            console.error('Error retrieving public keys:', err);
            throw new Error('Error retrieving public keys');
        }
    }
};

const retryWithDelay = (operation, retries = 5, delay = 7500) => new Promise((resolve, reject) => {
  return operation()
    .then(resolve)
    .catch((err) => {
      if (retries > 0) {
        return wait(delay)
          .then(retryWithDelay.bind(null, operation, retries - 1, delay))
          .then(resolve)
          .catch(reject);
      }

      return reject(err);
    });
});

const wait = (time_in_ms) => new Promise((resolve) => {
    setTimeout(() => resolve(), time_in_ms);
});

module.exports = database;
